from logging import debug, info
import boto3
from boto3.s3.transfer import S3Transfer


def upload_file(archive_bucket, artifact_name, file):
    debug(
        'Uploading artifact [%s] to bucket [%s] using archive [%s]'
        % (artifact_name, archive_bucket, file)
    )
    s3 = boto3.client('s3')
    client = S3Transfer(client=s3)
    client.upload_file(file, archive_bucket, artifact_name)
    info('File [%s] uploaded to bucket [%s]' % (artifact_name, archive_bucket))



'''
Lambda Gateway.

Usage:
  lgw gw-deploy [--verbose] --config-file=<cfg>
  lgw gw-undeploy [--verbose] --config-file=<cfg>
  lgw domain-add [--verbose] --config-file=<cfg>
  lgw domain-remove [--verbose] --config-file=<cfg>
  lgw lambda-deploy [--verbose] --config-file=<cfg> --lambda-file=<zip>
  lgw lambda-invoke [--verbose] --lambda-name=<name> [--payload=<json>]
  lgw lambda-delete [--verbose] --lambda-name=<name>
  lgw lambda-archive [--verbose] --config-file=<cfg>

Options:
  -h --help             Show this screen.
  --version             Show version.
  --verbose             Enable DEBUG-level logging.
  --config-file=<cfg>   Override defaults with these settings.
  --lambda-file=<zip>   Path to zip file with executable lambda code.
  --lambda-name=<name>  Name of the lambda to invoke or delete.
  --payload=<json>      Path to a file of type json with data to send with the lambda invocation.
'''

from os import path, makedirs
import json
from logging import info, debug, error
from everett.manager import ConfigManager, ConfigOSEnv, ConfigDictEnv
from docopt import docopt
from dotenv import dotenv_values, find_dotenv
from lgw.util import configure_logging
from lgw.version import __version__
from lgw import settings
from lgw.api_gateway import create_rest_api, delete_rest_api
from lgw.api_gateway_domain import add_domain_mapping, remove_domain_mapping
from lgw.lambda_util import deploy_function, invoke_function, delete_function
from lgw.lambda_bundle import build_lambda_archive


def handle_deploy_lambda(file, config):
    info('handle_deploy_lambda() called with file [%s]' % file)

    if not path.isfile(file):
        raise FileNotFoundError('ERROR: Lambda zip file not found at location: [%s]' % file)

    if not file.endswith('.zip'):
        raise FileNotFoundError('ERROR: Lambda file expected to be in ZIP format.')

    lambda_arn = deploy_function(
        file,
        config('aws_lambda_name'),
        config('aws_lambda_handler'),
        config('aws_lambda_execution_role_arn'),
        config('aws_lambda_connection_timeout'),
        config('aws_lambda_memory_size'),
        config('aws_lambda_runtime'),
        config('aws_lambda_archive_bucket'),
        config('aws_lambda_archive_key'),
        config('aws_lambda_description'),
        config('aws_lambda_vpc_subnets'),
        config('aws_lambda_vpc_security_groups'),
        config('aws_lambda_environment'),
        config('aws_lambda_tags'),
    )
    print(lambda_arn)
    info('Lambda [%s] created.' % config('aws_lambda_name'))
    return 1


def handle_lambda_archive(config):
    info(f'handle_lambda_archive() called.')
    addl_files = []
    if config('aws_lambda_archive_addl_files'):
        addl_files = [
            (x, y)
            for x, y in (
                line.split(',') for line in config('aws_lambda_archive_addl_files').split(';')
            )
        ]

    addl_packages = []
    if config('aws_lambda_archive_addl_packages'):
        addl_packages = config('aws_lambda_archive_addl_packages').split(',')

    context_dir = config('aws_lambda_archive_context_dir')
    if context_dir == '.':
        context_dir = path.abspath(context_dir)
        context_dir = f'{context_dir}/'

    if not path.exists(context_dir):
        raise FileNotFoundError(f'context dir not found: [context_dir]')

    bundle_dir = path.abspath(config('aws_lambda_archive_bundle_dir'))
    if not path.exists(bundle_dir):
        makedirs(bundle_dir)

    path_to_archive = build_lambda_archive(
        context_dir, bundle_dir, config('aws_lambda_archive_bundle_name'), addl_files, addl_packages
    )
    print(path_to_archive)
    info(f'lambda archive location: [{path_to_archive}]')


def handle_invoke_lambda(name, payload):
    info('handle_invoke_lambda() called for lambda [%s]' % name)
    if payload:
        result = invoke_function(name, payload)
    else:
        result = invoke_function(name)
    json_result = json.loads(result['Payload'].read().decode('utf-8'))
    print(json_result)
    info('Invocation completed for lambda [%s]' % name)
    return 1


def handle_delete_lambda(name):
    info('handle_delete_lambda() called for lambda [%s]' % name)
    delete_function(name)
    info('Lambda [%s] deleted.' % name)
    return 1


def handle_deploy_api_gateway(config):
    if config('aws_api_binary_types'):
        binary_types = config('aws_api_binary_types').split(',')

    response_models = {}
    if config('aws_api_response_models'):
        response_models = dict(
            item.split('=') for item in config('aws_api_response_models').split(';')
        )

    api_url = create_rest_api(
        config('aws_api_name'),
        config('aws_api_description'),
        binary_types,
        config('aws_lambda_name'),
        config('aws_api_resource_path'),
        config('aws_api_deploy_stage'),
        config('aws_api_lambda_integration_role'),
        response_models,
    )
    print(api_url)
    info('REST API URL: [%s]' % api_url)
    return 1


def handle_undeploy_api_gateway(config):
    delete_rest_api(config('aws_api_name'))
    info('API Gateway %s deleted.' % config('aws_api_name'))
    return 1


def handle_add_domain(config):
    api_name = config('aws_api_name')
    domain_name = config('aws_api_domain_name')
    base_path = config('aws_api_base_path')
    cert_arn = config('aws_acm_certificate_arn')
    deploy_stage = config('aws_api_deploy_stage')
    wait_until = config('aws_api_domain_wait_until_available')

    add_domain_mapping(api_name, domain_name, base_path, cert_arn, deploy_stage, wait_until)

    info(f'Domain name {domain_name} mapped to path {base_path}')
    info('HTTPS certificate validation may still be pending. Check here for status:')
    info('https://console.aws.amazon.com/apigateway/home?region=us-east-1#/custom-domain-names')

    return 1


def handle_remove_domain(config):
    api_name = config('aws_api_name')
    domain_name = config('aws_api_domain_name')
    base_path = config('aws_api_base_path')

    remove_domain_mapping(api_name, domain_name, base_path)

    info(f'Domain name {domain_name} unmapped from API {api_name}')
    return 1


def app(args, config):
    if args.get('gw-deploy'):
        return handle_deploy_api_gateway(config)
    if args.get('gw-undeploy'):
        return handle_undeploy_api_gateway(config)
    if args.get('domain-add'):
        return handle_add_domain(config)
    if args.get('domain-remove'):
        return handle_remove_domain(config)
    if args.get('lambda-deploy'):
        file = path.abspath(args.get('--lambda-file'))
        return handle_deploy_lambda(file, config)
    if args.get('lambda-invoke'):
        name = args.get('--lambda-name')
        payload = args.get('--payload', None)
        return handle_invoke_lambda(name, payload)
    if args.get('lambda-delete'):
        name = args.get('--lambda-name')
        return handle_delete_lambda(name)
    if args.get('lambda-archive'):
        return handle_lambda_archive(config)

    error('Unrecognized command.')


def load_config(config_file):
    # python-dotenv enables interpolation of values in config file
    # from the environment or elsewhere in the config file using
    # POSIX variable expansion
    local_conf = dotenv_values(find_dotenv())
    project_conf = dotenv_values(dotenv_path=config_file)
    config = ConfigManager(
        [
            ConfigOSEnv(),
            ConfigDictEnv(local_conf),
            ConfigDictEnv(project_conf),
            ConfigDictEnv(settings.defaults()),
        ]
    )
    return config


def main():
    args = docopt(__doc__, version=__version__)
    configure_logging(args.get('--verbose'))
    config_file = args.get('--config-file', None)
    if config_file:
        config_file = path.abspath(config_file)
    debug(f'Reading config from file {config_file}')

    config = load_config(config_file)

    app(args, config)


if __name__ == '__main__':
    main()



import docker
from io import BytesIO
import json
import ast
import tarfile
import os
import tempfile
from os.path import exists

from logging import debug, info, warn, error, getLogger, DEBUG

DOCKER_SOCKET_FILE = '/var/run/docker.sock'
DEFAULT_CODE_HOME = '/home/code/'
DEFAULT_VENV_HOME = '/home/venv/'
DEFAULT_OUTPUT_DIR = '/home/build/'
DEFAULT_PACKAGES = ['gcc', 'openssl-devel', 'bzip2-devel', 'libffi-devel', 'python37-pip']

DEFAULT_DOCKERIGNORE = [
    '**/.DS_Store',
    '.git/',
    '.gitignore',
    '.vscode/',
    '.idea/',
    '.gradle/',
    '.settings/',
    '**/*.pyc',
    '**/__pycache__/',
    '**/db.sqlite3',
    '*.sublime-project',
    '*.sublime-workspace',
    '*.retry',
    'vue-s3-dropzone/',
    '.pytest_cache/',
]


def build_lambda_archive(
    context_dir,
    lambda_archive_dir,
    lambda_archive_filename,
    addl_project_files=[],
    addl_yum_packages=[],
):
    if not exists(DOCKER_SOCKET_FILE):
        error(f'Docker listen socket not found at {DOCKER_SOCKET_FILE}')
        raise FileNotFoundError(
            f'Docker listen socket not found at {DOCKER_SOCKET_FILE}, is Docker running?'
        )

    info('Assembling Dockerfile.')
    dockerfile = create_dockerfile(lambda_archive_filename, addl_project_files, addl_yum_packages)
    debug(dockerfile)

    tag = 'lambda-bundle:latest'

    info(f'Building docker image based on files in {context_dir}')
    with tempfile.NamedTemporaryFile() as tmp:
        docker_context_file = create_docker_context(dockerfile, context_dir, tmp.name)
        cli = docker.APIClient(base_url=f'unix://{DOCKER_SOCKET_FILE}')
        for line in cli.build(fileobj=tmp, custom_context=True, encoding='gzip', tag=tag):
            print_progress(line)

    info('Running docker image to build lambda archive.')
    client = docker.from_env()
    container = client.containers.run(tag, command=f'/bin/sh', detach=True)

    info('Extracting lambda archive from running container.')
    bits, stat = container.get_archive(f'{DEFAULT_OUTPUT_DIR}/{lambda_archive_filename}')
    location = write_file_from_tar(bits, lambda_archive_dir, lambda_archive_filename)

    container.stop()

    return location


def create_dockerfile(archive_filename, addl_project_files, addl_yum_packages):
    yum_packages = ' '.join(set(DEFAULT_PACKAGES + addl_yum_packages))
    addl_files = ''
    for files in addl_project_files:
        addl_files += 'COPY %s %s\n' % files

    dockerfile = f'''
FROM lambci/lambda:build-python3.7 AS base

RUN yum makecache fast

RUN yum clean all && \
  yum update --assumeyes && \
  yum upgrade --assumeyes

RUN yum install  --assumeyes \
  {yum_packages}

ARG wkdir={DEFAULT_CODE_HOME}
RUN mkdir -p $wkdir

ARG venv={DEFAULT_VENV_HOME}
RUN mkdir -p $venv

ARG output={DEFAULT_OUTPUT_DIR}
RUN mkdir -p $output

WORKDIR $wkdir

{addl_files}

RUN python3 -m venv $venv
RUN source $venv/bin/activate && \
    pip3 install -U pip && \
    pip3 install -r requirements.txt

RUN echo "source $venv/bin/activate" > $HOME/.profile

RUN cd $wkdir && \
    zip -9 -r \
    $output/{archive_filename} . \
    --exclude '*/bin' '*dist-info*' '*__pycache__*' \
        '*.pyc' '*easy_install.py*' '*pip/*' \
            '*setuptools/*' '*pkg_resources/*'

RUN cd $venv/lib/python3.7/site-packages && \
    zip -9 -r -u \
    $output/{archive_filename} . \
    --exclude '*/bin' '*dist-info*' '*__pycache__*' \
        '*.pyc' '*easy_install.py*' '*pip/*' \
            '*setuptools/*' '*pkg_resources/*'

'''

    return dockerfile


def write_file_from_tar(data, dest, archive_filename):
    '''
    Extracts a single file named `archive_filename` from a tar
    file encoded in the stream `data` to the given location `dest`.
    Returns the new location of the extracted file.
    '''
    with tempfile.NamedTemporaryFile() as temp:
        for chunk in data:
            temp.write(chunk)
        temp.flush()
        os.fsync(temp)
        with tarfile.open(name=temp.name) as tf:

            logger = getLogger(__name__)
            if logger.isEnabledFor(DEBUG):
                for tinfo in tf.getnames():
                    debug(f'>    {tinfo}')

            tf.extract(archive_filename, path=dest)
    return f'{dest}/{archive_filename}'


def create_docker_context(dockerfile, context_directory, context_file):
    '''
    Collects the `context_directory` into a tar file
    along with the given `dockerfile`.
    Returns the filename containing the given context.
    '''

    def tar_exclude_filter(ti):
        for item in DEFAULT_DOCKERIGNORE:
            if item in ti.name:
                return None
        return ti

    with tarfile.open(context_file, 'w:gz') as tar:
        tar.add(
            context_directory,
            arcname=os.path.basename(context_directory),
            filter=tar_exclude_filter,
        )
        dockerfile_bytes = dockerfile.encode('utf8')
        info = tarfile.TarInfo(name='Dockerfile')
        info.size = len(dockerfile_bytes)
        tar.addfile(info, BytesIO(dockerfile_bytes))

        logger = getLogger(__name__)
        if logger.isEnabledFor(DEBUG):
            for tinfo in tar.getnames():
                debug(f'>    {tinfo}')

    return context_file


def print_progress(line):
    s = line.decode('utf8')
    if s:
        for l in s.split('\r\n'):
            if l.strip():
                try:
                    dd = ast.literal_eval(l.strip())
                except:
                    warn(l)
                else:
                    if 'stream' in dd:
                        ss = dd['stream']
                        if ss:
                            debug(ss.strip())
                    if 'error' in dd:
                        ee = dd['error']
                        if ee:
                            error(ee.strip())



from logging import debug, info, warn
import boto3
from lgw.api_gateway import lookup_api_gateway
from lgw.route53 import update_dns_a_record


def add_domain_mapping(
    api_name, domain_name, base_path, https_certificate_arn, deploy_stage, wait_for_completion
):
    api_client = boto3.client('apigateway')

    api_id = lookup_api_gateway(api_client, api_name)

    created = create_custom_domain_name(api_client, domain_name, https_certificate_arn)
    if wait_for_completion:
        while True:
            created = create_custom_domain_name(api_client, domain_name, https_certificate_arn)
            if created:
                break

    if created:
        debug(f'Adding base path mapping for {deploy_stage}')
        configure_base_path_mapping(api_client, api_id, domain_name, deploy_stage, base_path)

        cf_distribution = created['distributionDomainName']
        debug(f'Updating A record for {domain_name} with CF distribution {cf_distribution}')
        update_dns_a_record(domain_name, cf_distribution)


def remove_domain_mapping(api_name, domain_name, base_path):
    api_client = boto3.client('apigateway')

    api_id = lookup_api_gateway(api_client, api_name)

    response = api_client.get_base_path_mappings(domainName=domain_name)
    if response:
        for item in response['items']:
            print(item)

    api_client.delete_base_path_mapping(domainName=domain_name, basePath=base_path)

    api_client.delete_domain_name(domainName=domain_name)


def configure_base_path_mapping(api_client, api_id, domain_name, deploy_stage, base_path):
    response = None
    try:
        response = api_client.get_base_path_mapping(domainName=domain_name, basePath=base_path)
    except api_client.exceptions.NotFoundException:
        info(f'Base path mapping for {domain_name}:{base_path} does not exist.')

    if response:
        info(f'Base path mapping {domain_name}:{base_path} already exists.')
        api_client.delete_base_path_mapping(domainName=domain_name, basePath=base_path)

    response = api_client.create_base_path_mapping(
        domainName=domain_name, basePath=base_path, restApiId=api_id, stage=deploy_stage
    )

    if response['ResponseMetadata']['HTTPStatusCode'] == 201:
        info(f'Base path mapping created for {domain_name}:{base_path}')
    else:
        warn(f'Unable to create base path mapping for {domain_name}:{base_path}')


def create_custom_domain_name(api_client, domain_name, certificate_arn):
    response = None
    try:
        response = api_client.get_domain_name(domainName=domain_name)
    except api_client.exceptions.NotFoundException:
        info(f'Custom domain name {domain_name} does not exist.')

    if response:
        if response.get('domainNameStatus') == 'AVAILABLE':
            info(f'Domain name {domain_name} exists and is available.')
            return response

        if response.get('domainNameStatus') in ('UPDATING', 'PENDING'):
            info(
                'Domain name %s is in process: [%s] %s'
                % (
                    domain_name,
                    response.get('domainNameStatus'),
                    response.get('domainNameStatusMessage'),
                )
            )
            return

    response = api_client.create_domain_name(
        domainName=domain_name, certificateName=domain_name, certificateArn=certificate_arn
    )
    if response:
        cloudfront_distribution = response.get('distributionDomainName')
        info(f'domain name {domain_name} created, pointing at: {cloudfront_distribution}')
        return response



from logging import basicConfig, INFO, DEBUG, debug


def configure_logging(level=None):
    if not level:
        level = INFO
    else:
        level = DEBUG
    basicConfig(
        format='[%(asctime)s][%(levelname)s] %(message)s', datefmt='%Y/%m/%d %H:%M:%S', level=level,
    )



def defaults():
    return {
        'aws_region': 'us-east-1',
        'aws_api_name': '',
        'aws_api_description': '',
        'aws_api_resource_path': '{proxy+}',
        'aws_api_deploy_stage': '',
        'aws_api_lambda_integration_role': '',
        'aws_api_domain_name': '',
        'aws_api_base_path': '(none)',
        'aws_api_domain_wait_until_available': 'true',
        'aws_api_response_models': 'application/json=Empty',
        'aws_api_binary_types': '',
        'aws_acm_certificate_arn': '',
        'aws_lambda_name': '',
        'aws_lambda_description': '',
        'aws_lambda_handler': '',
        'aws_lambda_runtime': 'python3.7',
        'aws_lambda_connection_timeout': 30,
        'aws_lambda_memory_size': 3000,
        'aws_lambda_archive_bucket': '',
        'aws_lambda_archive_key': '',
        'aws_lambda_execution_role_arn': '',
        'aws_lambda_vpc_subnets': '',
        'aws_lambda_vpc_security_groups': '',
        'aws_lambda_environment': '',
        'aws_lambda_tags': '',
        'aws_lambda_archive_context_dir': '.',
        'aws_lambda_archive_bundle_dir': './build',
        'aws_lambda_archive_bundle_name': 'lambda-bundle.zip',
        'aws_lambda_archive_addl_files': '',
        'aws_lambda_archive_addl_packages': '',
    }
